import asyncio
import logging
from aionsq.http import NSQClient
from aionsq.lookup import NSQLookup
from aionsq.tcp import NSQReader, NSQWriter
from aionsq.util import parse_addr

__all__ = ['NSQ']


class NSQ:

    def __init__(self, addr=None, lookup=None, *, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        if addr and lookup:
            raise ValueError('mutually exclusive addr and lookup')
        if addr:
            addr = parse_addr(addr)
            if addr.proto == 'tcp':
                self.adapter = TCPAdapter(addr, loop=self.loop)
            elif addr.proto == 'http':
                self.adapter = HTTPAdapter(addr, loop=self.loop)
            else:
                raise ValueError('unkown proto %s' % addr.proto)
        elif lookup:
            self.adapter = LookupAdapter(lookup, loop=self.loop)
        else:
            raise ValueError('addr or lookup required')

    async def publish(self, topic, message):
        """Publish a message.

        Parameters:
            topic (str): the topic to publish to
            message (str): raw message bytes
        """
        client = await self.adapter.writer()
        response = await client.publish(topic, message)
        return response

    async def multi_publish(self, topic, messages):
        """Publish multiple messages in one roundtrip.

        Parameters:
            topic (str): the topic to publish to
            messages (list): raw message bytes
        """
        client = await self.adapter.writer()
        response = await client.multi_publish(topic, messages)
        return response

    def register(self, topic, channel, *, handler=None):
        """Register an handler that will be executed for each message received

        Parameters:
            handler (callable): the handler to add
        """
        if not handler:
            return lambda x: self.register(topic, channel, handler=x)

        # TODO is-it really working as expected?
        client = yield from self.adapter.reader(topic, channel)
        return client.register(handler)

    def close(self):
        self.adapter.close()

    async def wait_closed(self):
        await self.adapter.wait_closed()


class Adapter:

    async def reader(self, topic, channel):
        raise NotImplementedError()

    async def writer(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    async def wait_closed(self):
        raise NotImplementedError()


class HTTPAdapter(Adapter):

    def __init__(self, addr, *, loop=None):
        self.addr = addr
        self.loop = loop or asyncio.get_event_loop()
        self._writer = None

    async def reader(self, topic, channel):
        raise AttributeError('does not implements this')

    async def writer(self):
        if not self._writer:
            self._writer = NSQClient(self.addr, loop=self.loop)
        return self._writer

    def close(self):
        if self._writer:
            self._writer.close()

    async def wait_closed(self):
        if self._writer:
            await self._writer.wait_closed()


class LookupAdapter(Adapter):

    def __init__(self, addr, *, loop=None):
        self.log = logging.getLogger(__name__)
        self.loop = loop or asyncio.get_event_loop()
        self.client = NSQLookup(addr, loop=self.loop)
        self._writer = None

    async def reader(self, topic, channel):
        raise Exception()

    async def writer(self):
        if not self._writer:
            nodes = await self.client.nodes()
            node = nodes['producers'][0]
            addr = 'tcp://%s:%s' % (node['broadcast_address'], node['tcp_port'])
            self.log.info('connect to %s', addr)
            writer = NSQWriter(addr, loop=self.loop)
            await writer.start()
            self._writer = writer
        return self._writer

    def close(self):
        if self._writer:
            self._writer.close()

    async def wait_closed(self):
        if self._writer:
            await self._writer.wait_closed()


class TCPAdapter(Adapter):

    def __init__(self, addr, *, loop=None):
        self.addr = addr
        self.loop = loop or asyncio.get_event_loop()
        self._readers = {}
        self._writer = None

    async def reader(self, topic, channel):
        key = topic, channel
        if key not in self._readers:
            self._readers[key] = NSQReader(self.addr, loop=self.loop)
            await self._readers[key].start()
        return self._readers[key]

    async def writer(self):
        if not self._writer:
            self._writer = NSQWriter(self.addr, loop=self.loop)
            await self._writer.start()
        return self._writer

    def close(self):
        if self._writer:
            self._writer.close()
        for reader in self._readers.values():
            self._writer.close()

    async def wait_closed(self):
        if self._writer:
            await self._writer.wait_closed()
        for reader in self._readers.values():
            await self._writer.wait_closed()
