import asyncio
import struct
from .requests import RequestHandler, ok, exc, unwrapped_json


class HTTPHandler:

    def __init__(self, addr, *, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.req_handler = RequestHandler(addr, loop=self.loop)
        self._closed = asyncio.Event(loop=self.loop)

    def close(self):
        if not self._closed.is_set():
            self.req_handler.close()
            self._closed.set()

    async def wait_closed(self):
        await self._closed.wait()
        return True

    def __repr__(self):
        return '<%s(addr=%r)>' % (
            self.__class__.__name__,
            self.req_handler.addr
        )


class NSQClient(HTTPHandler):

    async def stats(self):
        """Return internal statistics.
        """
        params = {'format': 'json'}
        path = '/stats'
        response = await self.req_handler('GET', path, params=params)
        result = await response.json()
        return result['data']

    async def ping(self):
        """Monitoring endpoint.

        .. note::

            The only “unhealthy” state is if nsqd failed to write messages
            to disk when overflow occurred.
        """
        path = '/ping'
        response = await self.req_handler('GET', path)
        result = await ok(response)
        return result

    async def info(self):
        """Version information.
        """
        path = '/info'
        response = await self.req_handler('GET', path)
        result = await unwrapped_json(response)
        return result

    async def publish(self, topic, message):
        """Publish a message.

        Parameters:
            topic (str): the topic to publish to
            message (str): raw message bytes
        """
        params = {'topic': topic}
        path = '/pub'
        data = message.encode('utf-8')
        response = await self.req_handler('POST', path,
                                          params=params, data=data)
        result = await ok(response)
        return result

    async def multi_publish(self, topic, messages):
        """Publish multiple messages in one roundtrip.

        Parameters:
            topic (str): the topic to publish to
            messages (list): raw message bytes
        """
        assert messages, 'messages are required'
        params = {'topic': topic, 'binary': True}
        path = '/mpub'
        data = struct.pack('>l', len(messages))
        for message in messages:
            data += struct.pack('>l', len(message)) + message.encode('utf-8')
        response = await self.req_handler('POST', path,
                                          params=params, data=data)
        result = await ok(response)
        return result

    async def create_topic(self, topic):
        """Create a topic.

        Parameters:
            topic (str): the topic to create
        """
        params = {'topic': topic}
        path = '/topic/create'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def delete_topic(self, topic):
        """Delete an existing topic (and all channels).

        Parameters:
            topic (str): the existing topic to delete
        """
        params = {'topic': topic}
        path = '/topic/delete'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def create_channel(self, topic, channel):
        """Create a topic.

        Parameters:
            topic (str): the existing topic
            channel (str): the channel to create
        """
        params = {'topic': topic, 'channel': channel}
        path = '/channel/create'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def delete_channel(self, topic, channel):
        """Delete an existing channel on an existing topic.

        Parameters:
            topic (str): the existing topic to delete
            channel (str): the existing channel to delete
        """
        params = {'topic': topic, 'channel': channel}
        path = '/channel/delete'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def empty_topic(self, topic):
        """Empty all the queued messages (in-memory and disk)
        for an existing topic.

        Parameters:
            topic (str): the existing topic to empty
        """
        params = {'topic': topic}
        path = '/topic/empty'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def empty_channel(self, topic, channel):
        """Empty all the queued messages (in-memory and disk)
        for an existing topic.

        Parameters:
            topic (str): the existing topic
            channel (str): the existing channel to empty
        """
        params = {'topic': topic, 'channel': channel}
        path = '/channel/empty'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def pause_topic(self, topic):
        """Pause message flow to all channels on an existing topic
        (messages will queue at topic).

        Parameters:
            topic (str): the existing topic
        """
        params = {'topic': topic}
        path = '/topic/pause'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def unpause_topic(self, topic):
        """Resume message flow to channels of an existing, paused, topic.

        Parameters:
            topic (str): the existing topic
        """
        params = {'topic': topic}
        path = '/topic/unpause'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def pause_channel(self, topic, channel):
        """Pause message flow to consumers of an existing channel
        (messages will queue at channel).

        Parameters:
            topic (str): the existing topic
            channel (str): the existing channel to pause
        """
        params = {'topic': topic, 'channel': channel}
        path = '/channel/pause'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)

    async def unpause_channel(self, topic, channel):
        """Resume message flow to consumers of an existing, paused, channel.

        Parameters:
            topic (str): the existing topic
            channel (str): the existing channel to pause
        """
        params = {'topic': topic, 'channel': channel}
        path = '/channel/unpause'
        response = await self.req_handler('POST', path, params=params)
        if response.status == 200:
            await response.read()
            return True
        await exc(response)
