import asyncio
from aiohttp import ClientSession, TCPConnector
from .exceptions import TopicNotFound, ChannelNotFound
from .exceptions import HTTPInternalServerError, HTTPNotFound
from aionsq.util import parse_addr

__all__ = ['RequestHandler']


class RequestHandler:

    def __init__(self, addr, *, loop=None):
        self.addr = parse_addr(addr, proto='http', host='127.0.0.1')
        self.loop = loop or asyncio.get_event_loop()
        connector = TCPConnector(verify_ssl=False)
        self.session = ClientSession(connector=connector, loop=self.loop)
        self._closed = False

    async def request(self, method, path, **kwargs):
        url = '%s%s' % (self.addr, path)
        response = await self.session.request(method, url, **kwargs)
        return response

    def close(self):
        if not self._closed:
            self.session.close()
            self.session = None
            self._closed = True

    __call__ = request


async def exc(response):
    if response.status == 404:
        result = await response.json()
        if result == {'message': 'TOPIC_NOT_FOUND'}:
            raise TopicNotFound()
        if result == {'message': 'CHANNEL_NOT_FOUND'}:
            raise ChannelNotFound()
        raise HTTPNotFound(reason=result)
    result = await response.text()
    raise Exception(result)


async def ok(response):
    """Response must be 200 and 'OK'
    """
    if response.status == 200:
        result = await response.text()
        return result in ('OK', '')
    if response.status == 500:
        result = await response.text()
        raise Exception(result)
    raise NotImplementedError()


async def unwrapped_json(response):
    if response.status == 200:
        result = await response.json()
        return result['data']
    if response.status == 500:
        result = await response.json()
        raise HTTPInternalServerError(reason=result['status_txt'])
    raise NotImplementedError()
