import asyncio
import logging
from .connection import Connection
from .exceptions import NotStartedError
from .message import Message
from weakref import WeakKeyDictionary, WeakSet


class TCPHandler:

    def __init__(self, addr, *, auth_secret=None, cert=None, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self._conn = Connection(addr,
                                cert=cert,
                                auth_secret=auth_secret,
                                loop=self.loop)
        self._started = False
        self._openened = asyncio.Lock(loop=self.loop)
        self.log = logging.getLogger(__name__)

    @property
    def started(self):
        return self._started

    async def start(self, **opts):
        if not self._started:
            await self._conn.start()
            response = await self._conn.connect(**opts)
            await self.start_warmup()
            await self._openened.acquire()  # must be the last stmt
            self._started = True
        return response

    async def start_warmup(self):
        pass

    def close(self):
        self.log.debug('close')

        if self._started:
            task = self.loop.create_task(self.pre_close())
            task.add_done_callback(self._report_errors)
            task.add_done_callback(lambda x: self.force_close())
            self._started = False

    def force_close(self):
        self.log.debug('force close')
        self._conn.close()
        self._openened.release()

    async def wait_closed(self):
        with (await self._openened):
            return True

    async def pre_close(self):
        return None

    def _report_errors(self, fut):
        try:
            fut.result()
        except Exception as error:
            self.log.exception(error)


class NSQWriter(TCPHandler):

    async def publish(self, topic, message):
        """Publish a message.

        Parameters:
            topic (str): the topic to publish to
            message (str): raw message bytes
        """
        if not self._started:
            raise NotStartedError()
        await self._conn.cmd('PUB', topic, body=message)
        return True

    async def multi_publish(self, topic, messages):
        """Publish multiple messages in one roundtrip.

        Parameters:
            topic (str): the topic to publish to
            messages (list): raw message bytes
        """
        assert isinstance(messages, (list, set))
        if not self._started:
            raise NotStartedError()
        await self._conn.cmd('MPUB', topic, body=messages)
        return True


class NSQReader(TCPHandler):

    def __init__(self, addr, topic, channel, *args, **kwargs):
        super().__init__(addr, *args, **kwargs)
        self.topic = topic
        self.channel = channel
        self._consumers = WeakKeyDictionary()
        self._msg_counter = 0
        self._msg_rdy = 0
        self._msgs = WeakSet()

    @property
    def messages(self):
        """Returns pending messages
        """
        return set(self._msgs)

    @property
    def consumers(self):
        """Returns consumers
        """
        return set(self._consumers.keys())

    async def start_warmup(self):
        self._conn.add_message_listener(self.import_message)
        await self._conn.cmd('SUB', self.topic, self.channel)
        await self.ask_messages(1)

    def register(self, message_handler):
        """Register an handler that will be executed for each message received

        Parameters:
            message_handler (callable): the handler to add
        """
        cb = asyncio.coroutine(message_handler)
        self._consumers.setdefault(message_handler, cb)
        return message_handler

    def unregister(self, message_handler):
        """Unregister the handler.

        Parameters:
            message_handler (callable): the handler to remove
        """
        self._consumers.pop(message_handler, None)
        return message_handler

    def import_message(self, frame):
        msg = Message(self._conn, frame)
        self.log.info('msg %s', msg)
        self._msgs.add(msg)
        for consumer in self._consumers.values():
            fut = self.loop.create_task(consumer(msg))
            fut.add_done_callback(self._report_errors)

        self._msg_rdy -= 1
        if self._msg_rdy <= 0:
            self.loop.create_task(self.ask_messages(10))

    async def ask_messages(self, count):
        await self._conn.cmd('RDY', count)
        self._msg_rdy = count

    async def pre_close(self):
        await self._conn.cmd('CLS')
        # TODO ensure that response == 'CLOSE_WAIT'
        return True
