import asyncio
import atexit
import json
import logging
import socket
import ssl
from .exceptions import AuthFailedError
from .exceptions import IdentifyFailedError
from .exceptions import ServerError
from .handlers import DeflateHandler, SnappyHandler
from .helpers import prepare_cmd, to_size, parse_frame
from aionsq.util import parse_addr


class Connection:

    def __init__(self, addr, *, auth_secret=None, cert=None, loop=None):
        self.addr = parse_addr(addr, host='127.0.0.1', proto='tcp')
        self.auth_secret = auth_secret
        self.cert = cert
        self.log = logging.getLogger(__name__)
        self.loop = loop or asyncio.get_event_loop()
        self._started = False
        self.sock = None
        self.z = None
        self.message_listeners = set()

        self.max_bytes = 16 * 1024
        self._buf = bytearray()

    def add_message_listener(self, callback):
        self.message_listeners.add(callback)

    async def start(self):
        if not self._started:
            infos = await self.loop.getaddrinfo(*self.addr,
                                                family=socket.AF_INET,
                                                type=socket.SOCK_STREAM)
            *opts, _, addr = infos[0]
            self.sock = socket.socket(*opts)
            self.sock.settimeout(0)
            resp = await self.loop.sock_connect(self.sock, addr)
            atexit.register(self.close)
            self.log.error(resp)
            self.sock.send(b'  V2')
            self._started = True
            self.log.info('started')

    def close(self):
        self.sock.close()

    async def connect(self, **opts):
        from aionsq import __version__
        opts.setdefault('user_agent', 'aionsq/%s' % __version__)
        opts.setdefault('client_id', socket.getfqdn())
        opts.setdefault('hostname', socket.gethostname())
        opts.setdefault('feature_negotiation', True)
        if 'deflate' in opts:
            opts.setdefault('deflate_level', 6)
        data = json.dumps(opts)
        await self.cmd('IDENTIFY', body=data)
        features = await self.json()
        self.log.info('features %s', features)
        self.max_bytes = features['output_buffer_size']
        if features['tls_v1']:
            await self.upgrade_tls_v1(self.cert)
        if features['deflate']:
            await self.upgrade_deflate(6)
        if features['snappy']:
            await self.upgrade_snappy()
        if features['auth_required']:
            if not self.auth_secret:
                self.log.warning('auth required but auth_secret not set')
            auth_secret = self.auth_secret or '-'
            await self.cmd('AUTH', body=auth_secret)
            status = await self.json()
            self.log.info('auth status %s' % status)

        # TODO activate async now !!!()
        self.make_async()
        return features

    def make_async(self):
        self.loop.add_reader(self.sock.fileno(), self._read_ready)

    def bufferize_chunk(self, chunk):
        if self.z:
            chunk = self.z.decompress(chunk)
        self._buf += chunk

    async def upgrade_tls_v1(self, cert=None):
        self.log.info('upgrade tls')
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        if isinstance(cert, (list, tuple)):
            ctx.load_cert_chain(*cert)
        if isinstance(cert, dict):
            ctx.load_cert_chain(**cert)
        if isinstance(cert, str):
            ctx.load_cert_chain(cert)

        sock = ctx.wrap_socket(self.sock,
                               server_side=False,
                               do_handshake_on_connect=False)

        while True:
            try:
                sock.do_handshake()
            except ssl.SSLWantReadError:
                continue
            except ssl.SSLWantWriteError:
                continue
            break
        self.sock = sock
        await self.ok()

    async def upgrade_deflate(self, level):
        self.log.info('upgrade deflate')
        self.z = DeflateHandler(level)
        if len(self._buf):
            self._buf[:] = self.z.decompress(self._buf)
            self.log.warning('decompressed %s', self._buf)
        await self.ok()
        return

    async def upgrade_snappy(self):
        self.log.info('upgrade snappy')
        self.z = SnappyHandler()
        if len(self._buf):
            self._buf[:] = self.z.decompress(self._buf)
            self.log.warning('decompressed %s', self._buf)
        await self.ok()
        return

    async def cmd(self, cmd, *params, body=None):
        msg = prepare_cmd(cmd, *params, body=body)
        self.log.info('send %s', msg)
        await self.write(msg)

    async def write(self, msg):
        self.log.debug('write %s', msg)
        if self.z:
            msg = self.z.compress(msg)

        await self.loop.sock_sendall(self.sock, msg)

    def notify_listeners(self, frame):
        if frame.type == 0 and frame.body == '_heartbeat_':
            self.loop.create_task(self.send('NOP'))
            return
        elif frame.type == 0:
            self.log.debug('got %s', frame)
            return
        elif frame.type == 1 and frame.error in ('E_FIN_FAILED',
                                                 'E_REQ_FAILED',
                                                 'E_TOUCH_FAILED'):
            self.log.warning('failure %s', frame)
            return
        elif frame.type == 1:
            self.log.error('failure %s', frame)
            raise exc(frame)
        elif frame.type == 2:
            for listener in self.message_listeners:
                listener(frame)
            return
        raise NotImplementedError()

    async def json(self):
        data = await self.read()
        data = json.loads(data)
        self.log.debug('read json %s', data)
        return data

    async def ok(self):
        data = await self.read()
        return data == b'OK'

    def _read_ready(self):
        try:
            mbytes = self.max_bytes
            data = self.sock.recv(mbytes)
            self._data_received(data)
        except (BlockingIOError, InterruptedError, ssl.SSLWantReadError):
            return
        except Exception as error:
            self.log.exception(error)
            self.close()

    def _data_received(self, data):
        self.bufferize_chunk(data)
        while len(self._buf) > 4:
            data = self._buf[:4]
            size = to_size(data)
            self.log.debug('read size %s', size)
            if len(self._buf) < (4 + size):
                return

            data, self._buf[:4 + size] = self._buf[4:4 + size], []
            self.log.debug('read data %s', data)
            frame = parse_frame(data)
            self.notify_listeners(frame)

    async def _read(self, size):
        while len(self._buf) < size:
            chunk = bytearray()
            while True:
                try:
                    mbytes = self.max_bytes
                    chunk += await self.loop.sock_recv(self.sock, mbytes)
                except ssl.SSLWantReadError:
                    continue
                else:
                    break
            self.log.debug('read %s', chunk)
            self.bufferize_chunk(chunk)
        data, self._buf[:size] = self._buf[:size], []
        return data

    async def read(self):
        block = await self._read(4)
        size = to_size(block)
        self.log.debug('read size %s', size)
        data = await self._read(size)
        self.log.debug('read data %s', data)
        frame = parse_frame(data)
        self.log.debug('read frame %s', frame)
        if frame.type == 0:
            return frame.body
        if frame.type == 1:
            # fail fast
            raise exc(frame)
        # unexcepted
        raise Exception(frame)


def exc(frame):
    if frame.error == 'E_AUTH_FAILED':
        return AuthFailedError(frame.body)
    if frame.error == 'E_IDENTIFY_FAILED':
        return IdentifyFailedError(frame.body)
    return ServerError(frame)
