import struct
from collections import namedtuple
from .exceptions import MessageError

RawResponse = namedtuple('RawResponse', 'type body')
RawError = namedtuple('RawError', 'type error body')
RawMessage = namedtuple('RawMessage', 'type body timestamp attempts id')


def to_size(data):
    size, *_ = struct.unpack('>l', data)
    return size


def parse_frame(data):
    frame_type, *_ = struct.unpack('>l', data[:4])
    try:
        if frame_type == 0:
            return RawResponse(frame_type, data[4:].decode('utf-8'))
        elif frame_type == 1:
            error, body = data[4:].decode('utf-8').split(None, 1)
            return RawError(frame_type, error, body)
        elif frame_type == 2:
            timestamp = struct.unpack('>q', data[4:12])[0]
            attempts = struct.unpack('>h', data[12:14])[0]
            message_id = data[14:30].decode('utf-8')
            body = data[30:].decode('utf-8')
            return RawMessage(frame_type, body, timestamp,
                              attempts, message_id)
    except UnicodeDecodeError as error:
        raise MessageError(data) from error


def prepare_cmd(cmd, *params, body=None):
    output = bytearray()
    output += scalar(cmd)

    for param in params:
        output += b' '
        output += scalar(param)
    output += b'\n'

    if body is not None:
        if isinstance(body, (list, set)):
            buf = bytearray()
            buf += scalar_len(body)
            for o in body:
                a = scalar(o)
                buf += scalar_len(a)
                buf += a
        else:
            buf = scalar(body)
        output += scalar_len(buf)
        output += buf
    return bytes(output)


def scalar(obj):
    orig = obj
    if isinstance(obj, str):
        return bytearray(obj, encoding='utf-8')
    if isinstance(obj, (int, float)):
        return bytearray('%s' % obj, encoding='utf-8')
    if isinstance(obj, bytes):
        return bytearray(obj)
    if isinstance(obj, bytearray):
        return obj.copy()

    msg = "Don't know how to handle this %r" % orig.__class__.__name__
    raise NotImplementedError(msg)


def scalar_len(obj):
    return struct.pack('>l', len(obj))
